#include <iostream.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#include "Server.H"

static char *errmsg[] ={ "Unspecified error", "Unspecified error",
			 "Not implemented", "Obsolete call",
			 "Invalid password", "String too long",
			 "Login first", "Login disallowed",
			 "Conference zero", "Undefined conference",
			 "Undefined person", "Access denied",
			 "Permission denied", "Not member",
			 "No such text", "Text zero",
			 "No such local text", "Local text zero",
			 "Bad name", "Index out of range",
			 "Conference exists", "Person exists",
			 "Secret public", "Letterbox",
			 "Ldb error", "Illegal misc",
			 "Illegal info type", "Already recipient",
			 "Already comment", "Already footnote",
			 "Not recipient", "Not comment",
			 "Not footnote", "Recipient limit",
			 "Comment limit", "Footnote limit",
			 "Mark limit", "Not author",
			 "No connect", "Out of memory",
			 "Server is crazy", "Client is crazy",
			 "Undefined session", "Regexp error",
			 "Not marked", "Temporary failure",
			 "Long array", "Anonymous rejected",
			 "Illegal aux item", "Aux item permission",
			 "Unknown async", "Internal error",
			 "Feature disabled", "Message not sent",
			 "Invalid membership type" };

/* =======
   Methods
   ======= */

int Server::Command (int comm, ...)
{
  char buf[1000];
  va_list varg;
  va_start (varg, comm);
  
  switch (comm)
    {
    case Server_InitProtocol:
      {
	char *id;
	id = va_arg (varg, char *);
	sprintf (buf, "A%dH%s\n", strlen (id), id);
	write (fd, buf, strlen (buf));
	
	char buf[7];
	read (fd, buf, 7);
      }
      return 0;
      
    case Server_GetText:
      {
	int text, start, end;
	text = va_arg (varg, int);
	start = va_arg (varg, int);
	end = va_arg (varg, int);
	sprintf (buf, "%d 25 %d %d %d\n", ++seq, text, start, end);
	write (fd, buf, strlen (buf));
      }
      break;
      
    case Server_Login:
      {
	int pers;
	char *pass;
	char vis;
	pers = va_arg (varg, int);
	pass = va_arg (varg, char *);
	vis = va_arg (varg, int);
	sprintf (buf, "%d 62 %d %dH%s %d\n", ++seq, pers, strlen (pass),
		 pass, vis);
	write (fd, buf, strlen (buf));
      }
      break;
      
    case Server_LookupZName:
      {
	char *name;
	char pers, conf;
	name = va_arg (varg, char *);
	pers = va_arg (varg, int);
	conf = va_arg (varg, int);
	sprintf (buf, "%d 76 %dH%s %d %d\n", ++seq, strlen (name), name,
		 pers, conf);
	write (fd, buf, strlen (buf));
      }
      break;
      
    case Server_GetUconfStat:
      {
	int conf;
	conf = va_arg (varg, int);
	sprintf (buf, "%d 78 %d\n", ++seq, conf);
	write (fd, buf, strlen (buf));
      }
      break;
      
    case Server_GetConfStat:
      {
	int conf;
	conf = va_arg (varg, int);
	sprintf (buf, "%d 91 %d\n", ++seq, conf);
	write (fd, buf, strlen (buf));
      }
      break;
      
    case Server_LocalToGlobal:
      {
	int conf, first, no;
	conf = va_arg (varg, int);
	first = va_arg (varg, int);
	no = va_arg (varg, int);
	sprintf (buf, "%d 103 %d %d %d\n", ++seq, conf, first, no);
	write (fd, buf, strlen (buf));
      }
      break;

    default:
      cerr << "Panic on the ship! Internal command unknown!" << endl;
      exit (EXIT_FAILURE);
    }
  
  va_end (varg);
  
  return 0;
}

Token *Server::Read ()
{
  char chr;
  char buf[1000];
  
  while (1)
    {
      read (fd, &chr, 1);
      switch (chr)
	{
	case '=':
	  {
	    Token *token = NULL;
	    do
	      {
		read (fd, &chr, 1);
	      }
	    while (chr != ' ' && chr != '\n');
	    if (chr != '\n') token = new Token (fd, 0);
	    return token;
	  }
	  
	case ':':
	  do
	    {
	      read (fd, &chr, 1);
	    }
	  while (chr != '\n');
	  break;
	  
	case '%':
	  {
	    int cntr = 0;
	    do
	      {
		buf[cntr] = chr;
		read (fd, &chr, 1);
	      }
	    while (chr != '\n');
	    int err = atoi (buf);
	    cerr << "Error from server: ";
	    if (err > 1 && err < 55)
	      {
		cerr << errmsg[err];
	      }
	    strtok (buf, "\n");
	    cerr << " [" << buf << "]" << endl;
	    exit (EXIT_FAILURE);
	  }
	  
	default:
	  cerr << "Chaotic session: " << chr;
	  do
	    {
	      read (fd, &chr, 1);
	      cerr << chr;
	    }
	  while (chr != '\n');
	  exit (EXIT_FAILURE);
	}
    }
}

/* ==========================
   Constructor and destructor
   ========================== */

Server::Server (char *host, unsigned short port)
{
  int tmp;
  struct hostent *he;
  struct linger lng;
  struct sockaddr_in server;
  
  seq = 0;
  
  he = gethostbyname (host);
  if (he == NULL)
    {
      perror ("gethostbyname failed");
      exit (EXIT_FAILURE);
    }
  
  memset ((char *) &server, 0, sizeof (server));
  memcpy ((char *) &server.sin_addr, he->h_addr, he->h_length);
  server.sin_family = he->h_addrtype;
  server.sin_port = htons (port);
  
  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0)
    {
      perror ("Can't create socket");
      exit (EXIT_FAILURE);
    }
  
  lng.l_linger = 0;
  lng.l_onoff = 0;
  
  if (setsockopt (fd, SOL_SOCKET, SO_LINGER, &lng, sizeof(lng)) < 0)
    { 
      perror("Can't turn off linger");
      exit (EXIT_FAILURE);
    }
  
  tmp = 0;
  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof (tmp)) < 0)
    { 
      perror("Can't turn off reuse of local addr");
      exit (EXIT_FAILURE);
    }
  
  tmp = 1;
  if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &tmp, sizeof (tmp)) < 0)
    { 
      perror("Can't turn on KEEPALIVE");
      exit (EXIT_FAILURE);
    }
  
  if (connect (fd, (struct sockaddr *) &server, sizeof (server)) < 0)
    { 
      perror ("Could not connect");
      close (fd);
      exit (EXIT_FAILURE);
    }
}

Server::~Server ()
{
  close (fd);
}
